"""This module contains all the utility functions and classes used in the carwatch package."""
from carwatch.utils.click_helper import Condition, validate_mail_input, validate_saliva_distances, validate_subject_path
from carwatch.utils.study import Study
from carwatch.utils.utils import assert_file_ending, assert_is_dir, sanitize_str_for_tex, tex_to_pdf, write_to_file

__all__ = [
    "assert_is_dir",
    "assert_file_ending",
    "write_to_file",
    "tex_to_pdf",
    "sanitize_str_for_tex",
    "Study",
    "Condition",
    "validate_subject_path",
    "validate_mail_input",
    "validate_saliva_distances",
]



"""Helper functions for click."""
import re
from pathlib import Path

import click

from carwatch.utils.utils import assert_file_ending


class Condition(click.Option):
    """Helper class that displays options as prompt depending on flag options previously set by user."""

    def __init__(self, *args, **kwargs):
        """Display options as prompt depending on flag options previously set by user.

        To invoke this feature for an option, specify ``cls=Condition`` and ``pos_condition="conditional_option"``
        for positive relations and ``neg_condition="conditional_option"`` for negative relations.
        Note that ``conditional_option`` need to be either a flag or a bool.

        Parameters
        ----------
        condition: str
            name of a previously prompted variable that decides whether the current option is needed or not
        is_positive: bool
            ``True`` when the relation between current and conditional variable are positive,
            i.e., prompt for current option is shown when ``condition`` is ``True``
            ``False`` when the current option should be hidden if ``condition`` is ``True``

        """
        if "pos_condition" in kwargs:
            self.condition = kwargs.pop("pos_condition")
            self.is_positive = True
        else:
            self.condition = kwargs.pop("neg_condition")
            self.is_positive = False
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        is_condition = ctx.params[self.condition]

        if not is_condition:
            if self.is_positive:
                self.prompt = None
            else:
                self.required = True

        if is_condition:
            if not self.is_positive:
                self.prompt = None
            else:
                self.required = True

        return super().handle_parse_result(ctx, opts, args)


class NumericChoice(click.Option):
    """Helper class that displays options as prompt depending on number selection previously set by user."""

    def __init__(self, *args, **kwargs):
        """Display options as prompt depending on numeric options previously set by user.

        To invoke this feature for an option, specify ``cls=NumericChoice`` and ``pos_condition="conditional_option"``
        for positive relations and ``neg_condition="conditional_option"`` for negative relations.
        Note that ``conditional_option`` need to be either a flag or a bool.

        Parameters
        ----------
        option: str
            name of a previously prompted variable from a fixed int range
        option_map: dict
            defines whether current prompt is displayed for every possible value of the option parameter

        """
        self.option = kwargs.pop("chosen_number")
        self.option_map = kwargs.pop("option_map")
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        chosen_number = ctx.params[self.option]
        is_positive = self.option_map[chosen_number]
        if is_positive:
            self.required = True
        else:
            self.prompt = None
        return super().handle_parse_result(ctx, opts, args)


def validate_subject_path(ctx, param, value):  # pylint:disable=unused-argument # noqa: C901
    if value:
        try:
            assert_file_ending(Path(value), [".csv", ".txt"])
        except ValueError as e:
            raise click.BadParameter(str(e))
    return value


def validate_mail_input(ctx, param, value):  # pylint:disable=unused-argument # noqa: C901
    if value:
        email_regex = r"[^@]+@[^@]+\.[^@]+"  # pattern <...>@<...>.<...>
        if not re.fullmatch(email_regex, value):
            raise click.BadParameter(f"{value} is not an email address!")
    return value


def validate_saliva_distances(ctx, param, value):  # pylint:disable=unused-argument # noqa: C901
    if value:
        value = value.replace(" ", "")  # trim spaces
        if "," in value:
            distances = value.split(",")
            for dist in distances:
                if not dist.isdigit():
                    raise click.BadParameter("Saliva distances need to be comma-separated integers!")
        else:
            if not value.isdigit():
                raise click.BadParameter("Saliva distance needs to be an integer!")
    return value


def get_file_name(suffix: str) -> str:
    """Get file name from click context.

    Parameters
    ----------
    suffix: str
        suffix to append to file name

    """
    c = click.get_current_context()
    study_name = c.params["study_name"]
    return f"{study_name}{suffix}"



"""Module that contains the :class:`~carwatch.utils.study.Study` class."""
import sys
from pathlib import Path
from typing import Optional, Sequence, Union

import pandas as pd

from carwatch.utils.utils import assert_file_ending


class Study:
    """Class that represents a study."""

    MAX_NUM_SUBJECTS = 999  # maximal amount of subjects representable with EAN8

    def __init__(
        self,
        study_name: str,
        num_days: int,
        num_samples: int,
        num_subjects: Optional[int] = None,
        subject_path: Optional[Union[str, Path]] = None,
        subject_column: Optional[str] = "subject",
        subject_prefix: Optional[str] = None,
        has_evening_sample: bool = False,
        start_sample_from_zero: bool = False,
    ):
        """Class that represents a study.

        Parameters
        ----------
        study_name: str
            Study name that will be printed on the labels
        num_days: int
            Number of days the study will be conducted
        num_samples: int
            *Total* number of biomarker samples per day (including a potential evening sample!)
        num_subjects: int, optional
            Number of subjects in the study
        subject_path: str or :class:`~pathlib.Path`, optional
            Path to a tabular list of all subjects
        subject_column: str, optional
            Only used when ``subject_path`` is set, specifies the column name containing the subject ids,
            default value is ``"subject"``.
        subject_prefix: str, optional
            Add prefix to participant number (e.g., "VP_")
        has_evening_sample: bool, optional
            Whether a biomarker sample in the evening is also collected, default is ``False``
        start_sample_from_zero: bool, optional
            Whether the ID of the first biomarker sample is 0, default is ``False``
            When set to ``False``, ID starts from 1

        """
        self.study_name = study_name
        self.num_days = num_days
        self.num_samples = num_samples
        self.subject_path = subject_path
        if self.subject_path:
            self.subject_ids = self._determine_subject_ids(subject_path, subject_column)
            self.num_subjects = len(self.subject_ids)
        elif num_subjects:
            self.num_subjects = num_subjects
            self.subject_ids = [str(i) for i in range(1, self.num_subjects + 1)]
        else:
            raise ValueError(
                "Subject number unknown! Specification of either `num_subjects` or `subject_data` required!"
            )
        if self.num_subjects > Study.MAX_NUM_SUBJECTS:
            raise ValueError(f"Sorry, studies with more than {Study.MAX_NUM_SUBJECTS} participants are not supported!")
        # if subject_prefix:
        #     subject_prefix = subject_prefix
        self.subject_prefix = subject_prefix
        self.has_evening_sample = has_evening_sample
        self.start_sample_from_zero = start_sample_from_zero

    @staticmethod
    def _determine_subject_ids(subject_path: Union[str, Path], subject_column: str) -> Optional[Sequence[str]]:
        """Extract the IDs of the study participants.

        The IDs uf the study participants are extracted depending on the content of the subject data file.
        It is assumed that the subject data contains comma-separated tabular data with a header row.

        Parameters
        ----------
        subject_path : str or :class:`~pathlib.Path`
            Path to either a `*.csv` or `*.txt` file that contains the tabular participant data,
            assumes that every row corresponds to one subject
        subject_column : str
            name of the column that contains the subject ids

        Returns
        -------
        list
            list with subject IDs

        """
        subject_path = Path(subject_path)
        try:
            if assert_file_ending(subject_path, [".csv", ".txt"]):
                subject_data = pd.read_csv(subject_path)
                subject_ids = subject_data[subject_column]
                return subject_ids.to_list()
            return None
        # _assert_file_ending throws value error if file has wrong ending/doesn't exist
        except ValueError as e:
            print(e)
            sys.exit(1)

    def get_subject_name(self, subject_index: int) -> str:
        """Return the name of a subject.

        Parameters
        ----------
        subject_index : int
            index of the subject

        Returns
        -------
        str
            name of the subject

        """
        if self.subject_path:
            # no prefix or padding used
            return self.subject_ids[subject_index]
        # use padding and prefix if specified
        subject_name_padding = len(str(self.num_subjects))  # length of zero-padding depending on subject count
        subject_name = f"{subject_index:0{subject_name_padding}d}"
        if self.subject_prefix:
            subject_name = f"{self.subject_prefix}{subject_name}"
        return subject_name

    @property
    def subject_indices(self):
        """Return a list of subject indices.

        Returns
        -------
        list
            list of subject indices

        """
        return list(range(1, self.num_subjects + 1))

    @property
    def subject_names(self):
        """Return a list of subject names.

        Returns
        -------
        list
            list of subject names

        """
        return [self.get_subject_name(index - 1) for index in self.subject_indices]

    @property
    def day_indices(self):
        """Return a list of day indices.

        Returns
        -------
        list
            list of day indices

        """
        return list(range(1, self.num_days + 1))

    @property
    def sample_indices(self):
        """Return a list of sample indices.

        Returns
        -------
        list
            list of sample indices

        """
        return list(range(0, self.num_samples))



"""Example of how to generate QR codes for a study."""

from carwatch.qr_codes import QrCodeGenerator
from carwatch.utils import Study

if __name__ == "__main__":
    study = Study(
        study_name="ExampleStudy",
        num_days=3,
        num_subjects=15,
        num_samples=5,
        has_evening_sample=True,
    )
    generator = QrCodeGenerator(study=study, saliva_distances=[10, 10, 10], contact_email="dum@my.com")
    generator.generate(output_dir=".")



"""Example of how to generate labels for a study."""

from carwatch.labels import CustomLayout, LabelGenerator
from carwatch.utils import Study

if __name__ == "__main__":
    study = Study(
        study_name="ExampleStudy",
        num_days=3,
        num_subjects=15,
        num_samples=5,
        has_evening_sample=True,
    )
    generator = LabelGenerator(study=study, add_name=True, has_barcode=True)
    layout = CustomLayout(
        num_cols=3,
        num_rows=4,
        left_margin=3,
        right_margin=3,
        top_margin=2,
        bottom_margin=2,
        inter_col=0.2,
        inter_row=0.5,
    )
    generator.generate(output_dir=".", debug=True, layout=layout)



"""Module for generating QR codes to configure a study for the CARWatch app."""
import sys
from pathlib import Path
from typing import Sequence, Union

import qrcode
from qrcode.image.pil import PilImage

from carwatch.utils.study import Study
from carwatch.utils.utils import assert_is_dir, sanitize_str_for_qr


class QrCodeGenerator:
    """Class that is used to generate a QR-Code including all relevant study information for the CARWatch app.

    The QR-Code can be scanned by the CARWatch app to configure the app for a specific study.

    """

    def __init__(
        self,
        study: Study,
        saliva_distances: Union[int, Sequence[int]],
        contact_email: str,
        check_duplicates: bool = False,
        enable_manual_scan: bool = False,
    ):
        """Generate QR-Code encoding all relevant study information.

        Parameters
        ----------
        study: :obj:`~carwatch.utils.Study`
           Study object for which QR-Code will be created
        saliva_distances: int or list
            Time distances between morning samples in minutes ordered chronologically.
            For s morning saliva samples, s-1 saliva distances need to be specified.
            If the distances are constant between all morning samples,
        contact_email: str
            E-mail address that App log data will be shared with
        check_duplicates: bool
            If set to ``True``, the CAR Watch app will check and alert if a barcode has been checked twice
        enable_manual_scan: bool
            If set to ``True``, the CAR Watch app will allow manual barcode scanning apart from timed alarms

        """
        self.study = study
        self.saliva_distances = self._sanitize_distances(saliva_distances)
        self.contact = contact_email
        self.check_duplicates = check_duplicates
        self.enable_manual_scan = enable_manual_scan
        self.output_dir = None
        self.output_name = f"qr_code_{self.study.study_name}"

    def generate(self, output_dir: str = ".", output_name: Union[str, None] = None):
        """Generate a `*.png` file with QR code according to the properties of the created ``QrCodeGenerator``.

        Parameters
        ----------
        output_dir: str
            Path to the directory where the generated QR-Code will be stored
        output_name: str, optional
            Filename of the generated QR-Code

        """
        output_dir = Path(output_dir)
        try:
            assert_is_dir(output_dir)
        except ValueError as e:
            print(e)
            sys.exit(1)
        self.output_dir = output_dir
        if output_name:
            if output_name.endswith(".png"):
                # store without file ending
                output_name = output_name.rsplit(".", 1)[0]
            self.output_name = output_name
        qr_img = self._generate_qr_code()
        self._save_qr_img(qr_img)

    def _generate_qr_code(self) -> PilImage:
        """Generate a QR-Code with the study information.

        This method translates the study data into a QR-Code. The information is encoded in the following format:

        ``CARWATCH;N:<study_name>;D:<num_days>;S:<subject_list>;T:<saliva_distances>;
        E:<has_evening_sample>;M:<contact_email>;F:<check_duplicates>;``

        Returns
        -------
        :obj:`qrcode.image.pil.PilImage`
            QR-Code as PIL image

        """
        # sanitize inputs to prevent decoding issues
        forbidden = [";", ":", ","]
        study_name = sanitize_str_for_qr(self.study.study_name, forbidden)
        subject_names = []
        for name in self.study.subject_names:
            sanitized_name = sanitize_str_for_qr(name, forbidden)
            subject_names.append(sanitized_name)

        app_id = "CARWATCH"
        name_string = ",".join(str(dist) for dist in subject_names)
        distance_string = ",".join(str(dist) for dist in self.saliva_distances)

        # create encoding
        data = (
            f"{app_id};"
            f"N:{study_name};"
            f"D:{self.study.num_days};"
            f"S:{name_string};"
            f"T:{distance_string};"
            f"E:{int(self.study.has_evening_sample)};"
            f"M:{self.contact};"
            f"FD:{int(self.check_duplicates)};"
            f"FM:{int(self.enable_manual_scan)}"
        )

        # create qr code
        img = qrcode.make(data)
        return img

    def _save_qr_img(self, qr_img):
        img_location = self.output_dir.joinpath(f"{self.output_name}.png")
        qr_img.save(img_location)

    def _sanitize_distances(self, saliva_distances: Union[int, Sequence[int]]) -> Sequence[int]:
        # required number of saliva distances
        num_morning_samples = self.study.num_samples
        if self.study.has_evening_sample:
            # evening sample counts into total number of distances
            num_morning_samples = num_morning_samples - 1
        if isinstance(saliva_distances, int):
            return [saliva_distances] * (num_morning_samples - 1)
        if isinstance(saliva_distances, list):
            # all elements are ints
            if all(isinstance(dist, int) for dist in saliva_distances):
                # length is correct
                if len(saliva_distances) == num_morning_samples - 1:
                    return saliva_distances
                raise ValueError(
                    f"Incorrect number of saliva distances provided! "
                    f"Needs to be {num_morning_samples - 1}, as "
                    f"{num_morning_samples} morning samples will be taken."
                )
            raise ValueError("Invalid data detected in saliva distances! All values need to be integers!")
        raise ValueError("Saliva distances data type is invalid! Needs to be int or list of ints.")



"""Includes properties of predefined layouts for printing. All measurements need to be specified in mm."""
import abc


class PrintLayout:
    """Abstract class that represents a generic A4 label sheet."""

    __metaclass__ = abc.ABCMeta

    a4_height = 297
    a4_width = 210

    def get_label_height(self):
        """Return height of each label in mm.

        Returns
        -------
        float
            Height of each label in mm.

        """
        height_all_labels = self.a4_height - self.top_margin - self.bottom_margin - self.inter_row * (self.num_rows - 1)
        return height_all_labels / self.num_rows

    @property
    @abc.abstractmethod
    def num_cols(self) -> int:
        """Return number of distinct labels per column."""

    @num_cols.setter
    def num_cols(self, value):
        self.num_cols = value

    @property
    @abc.abstractmethod
    def num_rows(self) -> int:
        """Return number of distinct labels per row."""

    @num_rows.setter
    def num_rows(self, value):
        self.num_rows = value

    @property
    @abc.abstractmethod
    def left_margin(self) -> float:
        """Return offset between edge of sheet and first label to the left in mm."""

    @left_margin.setter
    def left_margin(self, value):
        self.left_margin = value

    @property
    @abc.abstractmethod
    def right_margin(self) -> float:
        """Return offset between edge of sheet and first label to the right in mm."""

    @right_margin.setter
    def right_margin(self, value: float):
        """Set offset between edge of sheet and first label to the right in mm.

        Parameters
        ----------
        value : float
            Offset between edge of sheet and first label to the right in mm.

        """
        self.right_margin = value

    @property
    @abc.abstractmethod
    def top_margin(self) -> float:
        """Return offset between edge of sheet and first label to the top in mm."""

    @top_margin.setter
    def top_margin(self, value: float):
        """Set offset between edge of sheet and first label to the top in mm.

        Parameters
        ----------
        value : float
            Offset between edge of sheet and first label to the top in mm.

        """
        self.top_margin = value

    @property
    @abc.abstractmethod
    def bottom_margin(self) -> float:
        """Return offset between edge of sheet and first label to the bottom in mm."""

    @bottom_margin.setter
    def bottom_margin(self, value: float):
        """Set offset between edge of sheet and first label to the bottom in mm.

        Parameters
        ----------
        value : float
            Offset between edge of sheet and first label to the bottom in mm.

        """
        self.bottom_margin = value

    @property
    @abc.abstractmethod
    def inter_col(self) -> float:
        """Return distance between each label along the columns in mm."""

    @inter_col.setter
    def inter_col(self, value: float):
        """Set distance between each label along the columns in mm.

        Parameters
        ----------
        value : float
            Distance between each label along the columns in mm.

        """
        self.inter_col = value

    @property
    @abc.abstractmethod
    def inter_row(self) -> float:
        """Return distance between each label along the rows in mm."""

    @inter_row.setter
    def inter_row(self, value: float):
        """Set distance between each label along the rows in mm.

        Parameters
        ----------
        value : float
            Distance between each label along the rows in mm.

        """
        self.inter_row = value


class AveryZweckformJ4791Layout(PrintLayout):
    """Class that represents the layout of `AveryZweckformJ4791` label sheets."""

    num_cols = 4
    num_rows = 12
    left_margin = 9.8
    right_margin = 9.8
    top_margin = 21.2
    bottom_margin = 21.2
    inter_col = 2.5
    inter_row = 0


class CustomLayout(PrintLayout):
    """Class that represents a user-defined label sheet layout."""

    num_cols = 0
    num_rows = 0
    left_margin = 0
    right_margin = 0
    top_margin = 0
    bottom_margin = 0
    inter_col = 0
    inter_row = 0

    def __init__(
        self,
        num_cols: int,
        num_rows: int,
        left_margin: float,
        right_margin: float,
        top_margin: float,
        bottom_margin: float,
        inter_col: float,
        inter_row: float,
    ):
        """Initialize a custom layout.

        Parameters
        ----------
        num_cols : int
            Number of distinct labels per column.
        num_rows : int
            Number of distinct labels per row.
        left_margin : float
            Offset between edge of sheet and first label to the left in mm.
        right_margin : float
            Offset between edge of sheet and last label to the right in mm.
        top_margin : float
            Offset between edge of sheet and first label to the top in mm.
        bottom_margin : float
            Offset between edge of sheet and last label to the bottom in mm.
        inter_col : float
            Distance between each label along the columns in mm.
        inter_row : float
            Distance between each label along the rows in mm.

        """
        self.num_cols = num_cols
        self.num_rows = num_rows
        self.left_margin = left_margin
        self.right_margin = right_margin
        self.top_margin = top_margin
        self.bottom_margin = bottom_margin
        self.inter_col = inter_col
        self.inter_row = inter_row



"""CARWatch: A Python package for improving the accuracy of cortisol awakening response sampling."""

__version__ = "0.1.1"

from carwatch import examples  # noqa: F401
from carwatch import labels  # noqa: F401

__all__ = ["labels"]



"""Utility functions for the carwatch package."""
import os
import subprocess
import warnings
from pathlib import Path
from subprocess import check_call
from typing import Optional, Sequence, Union


def assert_is_dir(path: Path) -> Optional[bool]:
    """Check if a path is a directory.

    Parameters
    ----------
    path : str
        path to check if it's a directory

    Returns
    -------
    ``True`` if ``path`` is a directory

    Raises
    ------
    ValueError
        if ``path`` is not a directory

    """
    # ensure pathlib
    file_name = Path(path)
    if not file_name.is_dir():
        raise ValueError(f"The path '{path}' is expected to be a directory, but it's not!")
    return True


def write_to_file(file: Path, content: str):
    """Write a given text to a file.

    If `file` doesn't exist it is created, otherwise, it is truncated.

    Parameters
    ----------
    file: :class:`~pathlib.Path`
        path to file that will be written
    content : str
        string that will be inserted to file

    """
    # ensure pathlib
    file_name = Path(file)
    with open(file_name, "w+", encoding="utf-8") as fp:
        fp.write(content)


def tex_to_pdf(output_dir: Path, tex_file: Union[str, Path]):
    """Run shell command to compile a tex file to a pdf document.

    Parameters
    ----------
    output_dir: path
        Path to directory where the output file will be stored
    tex_file: str or :class:`~pathlib.Path`
        Path to the `*.tex`-file that will be compiled

    Raises
    ------
    RuntimeError
        if ``pdflatex`` is not found

    """
    try:
        #  call pdflatex and suppress console output
        check_call(["pdflatex", f"-output-directory={output_dir}", tex_file], stdout=subprocess.DEVNULL, timeout=60)
    except FileNotFoundError as e:
        if os.name.startswith("win"):
            raise RuntimeError(
                "Apparently, you don't have Latex installed. "
                "Please install MikTex (from here: https://miktex.org/download), restart your computer, and try again."
            ) from e
        raise RuntimeError(
            "Apparently, you don't have Latex installed. "
            "Please install TexLive (from here: https://tug.org/texlive/) and try again."
        ) from e
    except subprocess.TimeoutExpired as e:
        # when pdflatex gets stuck, process will not end automatically
        raise RuntimeError(
            "Compilation aborted as it took too long. Please check your input parameters for plausibility."
        ) from e
    print(f"\nPDF created successfully and can be found here: {output_dir.absolute()}")


def assert_file_ending(path: Path, ending: Union[str, Sequence[str]]) -> bool:
    """Check if a path points to an existing file with a certain file ending.

    Parameters
    ----------
    path : :class:`~pathlib.Path`
        path to check if it's a directory
    ending : str or list of str
        file ending or a list of file endings that the file is expected to have

    Returns
    -------
    ``True`` if ``path`` has the given file ending

    Raises
    ------
    ValueError
        if ``path`` is not pointing to a file, or
        if the file extension is invalid

    """
    if path.is_file():
        if isinstance(ending, str):
            ending = [ending]
        for end in ending:
            if str(path).endswith(end):
                return True
        raise ValueError("The file has an invalid extension! It needs to be either a .csv or .txt file!")
    raise ValueError(f"The path is '{path}' is not an existing file!")


def sanitize_str_for_tex(string: str) -> str:
    r"""Escape special characters in a string to preserve them when compiled with pdflatex.

    The characters that are escaped are: `"&", "%", "$", "#", "_", "{", "}", "~", "^", "\"`.

    Parameters
    ----------
    string : str
       an arbitrary unescaped string

    Returns
    -------
    str
        Sanitized version of ``string`` with all characters of special meaning in latex escaped

    """
    escape_chars = ["&", "%", "$", "#", "_", "{", "}"]
    replace_chars = {"~": r"\textasciitilde", "^": r"\textasciicircum", r"\\": r"\textbackslash"}
    for c in escape_chars:
        string = string.replace(c, rf"\{c}")
    for c, val in replace_chars.items():
        string = string.replace(c, f"{val}")
    return string


def sanitize_str_for_qr(string: str, forbidden_chars: Sequence[str]) -> str:
    r"""Remove special characters in a string to prevent parsing problems when reading the data.

    Parameters
    ----------
    string : str
        an arbitrary unescaped string
    forbidden_chars : list
        characters that will be removed from the string

    Returns
    -------
    str
        Sanitized version of ``string`` with all forbidden characters removed

    """
    for c in forbidden_chars:
        if c in string:
            warnings.warn(f"Forbidden symbol `{c}` detected in input `{string}`, will be removed.")
        string = string.replace(c, "")
    return string
